from typing import Optional
from .. import cache

_USERNAME_KEY = "username"


def set_username(value: str) -> None:
    cache.store(_USERNAME_KEY, value)


def get_username() -> Optional[str]:
    return cache.get(_USERNAME_KEY)



"""Implements a simple persistent file-store cache.

Values set using this cache will persist across CLI command invocations.
"""

import os
from typing import Optional


def _cache_dir() -> str:
    """Hack to prevent circular imports issue with the constants module"""
    from .config import CIDC_WORKING_DIR

    return CIDC_WORKING_DIR


def _key_path(key: str) -> str:
    return os.path.join(_cache_dir(), key)


def store(key: str, value: str) -> None:
    """Persist a value across CLI commands."""
    cdir = _cache_dir()
    # Create the cache directory if it doesn't exist
    if not os.path.exists(cdir):
        os.mkdir(cdir)

    # Save the provided value in a file named key
    with open(_key_path(key), "w") as cache:
        cache.write(value)


def get(key: str) -> Optional[str]:
    """Try to get a value for the given key"""
    key_path = _key_path(key)

    # Check if the key exists
    if not os.path.exists(key_path):
        return None

    # Get the value
    with open(key_path, "r") as value:
        return value.read()



#!/usr/bin/env python
"""
Configuration file for pytest.
"""
__author__ = "Lloyd McCarthy"
__license__ = "MIT"

import sys
from os.path import abspath, dirname

import pytest
from click.testing import CliRunner

PACKAGE_PATH = abspath(dirname(dirname(__file__)))
sys.path.insert(0, PACKAGE_PATH)


@pytest.fixture
def runner():
    return CliRunner()



TEST_MANIFEST_ID: str = "test_upload"
TEST_PASSWORD: str = "password"
TEST_TRIAL_ID: str = "test_prism_trial_id"
TEST_USER: str = "test_user"

TEST_CLINICAL_URL_XLSX: str = "clinical_file1.xlsx"
TEST_CLINICAL_URL_CSV: str = "clinical_file2.csv"
TEST_MISC_DATA_URL1: str = "misc_data1.foo"
TEST_MISC_DATA_URL2: str = "misc_data2.bar"

TEST_METADATA_JSON: dict = {
    "shipments": [
        {"manifest_id": TEST_MANIFEST_ID},
        {"manifest_id": TEST_MANIFEST_ID + "2"},
    ],
    "participants": [
        {
            "cimac_participant_id": "CTTTPP1",
            "samples": [
                {"cimac_id": "CTTTPP101.00"},
                {"cimac_id": "CTTTPP102.00"},
            ],
        },
        {
            "cimac_participant_id": "CTTTPP2",
            "samples": [
                {"cimac_id": "CTTTPP201.00"},
                {"cimac_id": "CTTTPP202.00"},
            ],
        },
        {
            "cimac_participant_id": "CTTTPP3",
            "samples": [
                {"cimac_id": "CTTTPP301.00"},
                {"cimac_id": "CTTTPP302.00"},
                {"cimac_id": "CTTTPP303.00"},
            ],
        },
    ],
    "clinical_data": {
        "records": [
            {
                "clinical_file": {
                    "object_url": f"{TEST_TRIAL_ID}/clinical/{TEST_CLINICAL_URL_XLSX}",
                    "number_of_participants": 5,
                },
                "comment": "comment",
            },
            {
                "clinical_file": {
                    "object_url": f"{TEST_TRIAL_ID}/clinical/{TEST_CLINICAL_URL_CSV}",
                    "number_of_participants": 3,
                },
            },
        ],
    },
    "analysis": {
        "atacseq_analysis": [
            {
                "batch_id": "atacseq_analysis_batch",
                "report": {
                    "object_url": f"{TEST_TRIAL_ID}/atacseq/analysis/atacseq_analysis_batch/report.zip",
                },
                "records": [
                    {
                        "cimac_id": "CTTTPP101.00",
                        "aligned_sorted_bam": {
                            "object_url": f"{TEST_TRIAL_ID}/atacseq/CTTTPP101.00/analysis/aligned_sorted.bam",
                        },
                    },
                    {
                        "cimac_id": "CTTTPP201.00",
                        "aligned_sorted_bam": {
                            "object_url": f"{TEST_TRIAL_ID}/atacseq/CTTTPP201.00/analysis/aligned_sorted.bam",
                        },
                    },
                ],
            },
            {
                "batch_id": "atacseq_analysis_batch_2",
                "report": {
                    "object_url": f"{TEST_TRIAL_ID}/atacseq/analysis/atacseq_analysis_batch_2/report.zip",
                },
                "records": [
                    {
                        "cimac_id": "CTTTPP102.00",
                        "aligned_sorted_bam": {
                            "object_url": f"{TEST_TRIAL_ID}/atacseq/CTTTPP102.00/analysis/aligned_sorted.bam",
                        },
                    },
                ],
            },
        ],
        "tcr_analysis": {
            "batches": [
                {
                    "batch_id": "tcr_analysis_batch",
                    "report_trial": {
                        "object_url": f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch/report_trial.tar.gz",
                    },
                    "records": [
                        {
                            "cimac_id": "CTTTPP101.00",
                            "tra_clone": {
                                "object_url": f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch/CTTTPP101.00/tra_clone.csv",
                            },
                        },
                        {
                            "cimac_id": "CTTTPP201.00",
                            "tra_clone": {
                                "object_url": f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch/CTTTPP201.00/tra_clone.csv",
                            },
                        },
                    ],
                },
                {
                    "batch_id": "tcr_analysis_batch_2",
                    "report_trial": {
                        "object_url": f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch_2/report_trial.tar.gz",
                    },
                    "records": [
                        {
                            "cimac_id": "CTTTPP102.00",
                            "tra_clone": {
                                "object_url": f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch_2/CTTTPP102.00/tra_clone.csv",
                            },
                        },
                    ],
                },
            ],
        },
        "rna_analysis": {
            "level_1": [
                {
                    "cimac_id": "CTTTPP101.00",
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/rna/CTTTPP101.00/analysis/error.yaml",
                    },
                },
                {
                    "cimac_id": "CTTTPP201.00",
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/rna/CTTTPP201.00/analysis/error.yaml",
                    },
                },
                {
                    "cimac_id": "CTTTPP102.00",
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/rna/CTTTPP102.00/analysis/error.yaml",
                    },
                },
            ],
        },
        "wes_analysis": {
            "pair_runs": [
                {
                    "run_id": "CTTTPP101.00",  # tumor cimac_id
                    "normal": {"cimac_id": "CTTTPP10N.00"},  # cimac_id
                    "tumor": {"cimac_id": "CTTTPP101.00"},  # cimac_id
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes/CTTTPP101.00/analysis/error.yaml",
                    },
                },
                {
                    "run_id": "CTTTPP102.00",  # tumor cimac_id
                    "normal": {"cimac_id": "CTTTPP10N.00"},  # cimac_id
                    "tumor": {"cimac_id": "CTTTPP102.00"},  # cimac_id
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes/CTTTPP102.00/analysis/error.yaml",
                    },
                },
                {
                    "run_id": "CTTTPP201.00",  # tumor cimac_id
                    "normal": {"cimac_id": "CTTTPP20N.00"},  # cimac_id
                    "tumor": {"cimac_id": "CTTTPP201.00"},  # cimac_id
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes/CTTTPP201.00/analysis/error.yaml",
                    },
                },
            ],
        },
        "wes_analysis_old": {
            "pair_runs": [
                {
                    "run_id": "CTTTPP201.00",  # tumor cimac_id
                    "normal": {"cimac_id": "CTTTPP20N.00"},  # cimac_id
                    "tumor": {"cimac_id": "CTTTPP201.00"},  # cimac_id
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes/CTTTPP201.00/analysis/error.yaml",
                    },
                },
                {
                    "run_id": "CTTTPP102.00",  # tumor cimac_id
                    "normal": {"cimac_id": "CTTTPP10N.00"},  # cimac_id
                    "tumor": {"cimac_id": "CTTTPP102.00"},  # cimac_id
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes/CTTTPP102.00/analysis/error.yaml",
                    },
                },
            ],
        },
        "wes_tumor_only_analysis": {
            "runs": [
                {
                    "tumor": {"cimac_id": "CTTTPP101.00"},
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes_tumor_only/CTTTPP101.00/analysis/error.yaml",
                    },
                },
                {
                    "tumor": {"cimac_id": "CTTTPP102.00"},
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes_tumor_only/CTTTPP102.00/analysis/error.yaml",
                    },
                },
                {
                    "tumor": {"cimac_id": "CTTTPP201.00"},
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes_tumor_only/CTTTPP201.00/analysis/error.yaml",
                    },
                },
            ],
        },
        "wes_tumor_only_analysis_old": {
            "runs": [
                {
                    "tumor": {"cimac_id": "CTTTPP201.00"},
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes_tumor_only/CTTTPP201.00/analysis/error.yaml",
                    },
                },
                {
                    "tumor": {"cimac_id": "CTTTPP102.00"},
                    "error": {
                        "object_url": f"{TEST_TRIAL_ID}/wes_tumor_only/CTTTPP102.00/analysis/error.yaml",
                    },
                },
            ],
        },
    },
    "assays": {
        "cytof": [
            {
                "assay_run_id": "cytof_run",
                "batch_id": "cytof_batch",
                "astrolabe_analysis": {
                    "object_url": f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch/reports.zip",
                },
                "source_fcs": [
                    {
                        "object_url": f"{TEST_TRIAL_ID}/cytof/cytof_batch/source_0.fcs",
                    },
                ],
                "records": [
                    {
                        "cimac_id": "CTTTPP101.00",
                        "input_files": {
                            "processed_fcs": {
                                "object_url": f"{TEST_TRIAL_ID}/cytof/CTTTPP101.00/processed.fcs",
                            },
                        },
                        "output_files": {
                            "assignment": {
                                "object_url": f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch/CTTTPP101.00/assignment.csv",
                            },
                            "fcs_file": {
                                "object_url": f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch/CTTTPP101.00/source.fcs",
                            },
                        },
                    },
                    {
                        "cimac_id": "CTTTPP201.00",
                        "output_files": {
                            "assignment": {
                                "object_url": f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch/CTTTPP201.00/assignment.csv",
                            },
                            "fcs_file": {
                                "object_url": f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch/CTTTPP201.00/source.fcs",
                            },
                        },
                    },
                ],
            },
            {
                "assay_run_id": "cytof_run",
                "batch_id": "cytof_batch_2",
                "astrolabe_analysis": {
                    "object_url": f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch_2/reports.zip",
                },
                "records": [
                    {
                        "cimac_id": "CTTTPP102.00",
                        "output_files": {
                            "assignment": {
                                "object_url": f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch_2/CTTTPP102.00/assignment.csv",
                            },
                            "fcs_file": {
                                "object_url": f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch_2/CTTTPP102.00/source.fcs",
                            },
                        },
                    },
                ],
            },
            {
                "assay_run_id": "cytof_run_2",
                "batch_id": "cytof_batch_3",
                "records": [
                    {
                        "cimac_id": "CTTTPP301.00",
                    },
                ],
            },
        ],
        "misc_data": [
            {
                "files": [
                    {
                        "file": {
                            "object_url": f"{TEST_TRIAL_ID}/misc_data/{TEST_MISC_DATA_URL1}"
                        },
                        "description": "description",
                    },
                ],
            },
            {
                "files": [
                    {
                        "file": {
                            "object_url": f"{TEST_TRIAL_ID}/misc_data/{TEST_MISC_DATA_URL2}"
                        },
                    },
                ],
            },
        ],
        "olink": {
            "dummy_file": {
                "object_url": f"{TEST_TRIAL_ID}/olink/ALL object_urls",
            },
            # first priority, just this if exists
            "study": {
                "npx_file": {
                    "samples": [
                        "CTTTPP101.00",
                        "CTTTPP201.00",
                        "CTTTPP102.00",
                    ],  # cimac_ids
                    "object_url": f"{TEST_TRIAL_ID}/olink/study_npx.xlsx",
                },
                "dummy_file": {
                    "object_url": f"{TEST_TRIAL_ID}/olink/ALL object_urls",
                },
            },
            "batches": [
                {
                    "batch_id": "olink_batch",
                    "dummy_file": {
                        "object_url": f"{TEST_TRIAL_ID}/olink/batch_olink_batch/ALL object_urls",
                    },
                    # second priority, combines across batches
                    "combined": {
                        "npx_file": {
                            "samples": ["CTTTPP101.00"],
                            "object_url": f"{TEST_TRIAL_ID}/olink/batch_olink_batch/combined_npx.xlsx",
                        },
                    },
                    "records": [
                        {
                            # ignored
                            "files": {
                                "assay_npx": {
                                    "samples": ["CTTTPP101.00"],
                                    "object_url": f"{TEST_TRIAL_ID}/olink/batch_olink_batch/chip_0/assay_npx.xlsx",
                                },
                                "dummy_file": {
                                    "object_url": f"{TEST_TRIAL_ID}/olink/batch_olink_batch/chip_0/ALL object_urls",
                                },
                            },
                        },
                    ],
                },
                {
                    "batch_id": "olink_batch_2",
                    "records": [
                        {
                            # third priority, combines across batches
                            "files": {
                                "assay_npx": {
                                    "samples": ["CTTTPP201.00"],
                                    "object_url": f"{TEST_TRIAL_ID}/olink/batch_olink_batch_2/chip_0/assay_npx.xlsx",
                                },
                            },
                        },
                        {
                            # third priority, combines across batches
                            "files": {
                                "assay_npx": {
                                    "samples": ["CTTTPP102.00"],
                                    "object_url": f"{TEST_TRIAL_ID}/olink/batch_olink_batch_2/chip_1/assay_npx.xlsx",
                                },
                            },
                        },
                    ],
                },
            ],
        },
        "elisa": [
            {
                "assay_run_id": "elisa_batch",
                "assay_xlsx": {
                    "samples": ["CTTTPP101.00", "CTTTPP201.00"],
                    "object_url": f"{TEST_TRIAL_ID}/elisa/elisa_batch/assay.xlsx",
                },
            },
            {
                "assay_run_id": "elisa_batch_2",
                "assay_xlsx": {
                    "samples": ["CTTTPP102.00"],
                    "object_url": f"{TEST_TRIAL_ID}/elisa/elisa_batch_2/assay.xlsx",
                },
                "dummy_file": {
                    "object_url": f"{TEST_TRIAL_ID}/elisa/elisa_batch_2/ALL object_urls",
                },
            },
        ],
        "nanostring": [
            {
                "batch_id": "nanostring_batch",
                "data": {
                    "normalized": {
                        "object_url": f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/normalized_data.csv",
                    },
                },
                "runs": [
                    {
                        "run_id": "nanostring_batch_run",
                        "control_raw_rcc": {
                            "object_url": f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/nanostring_batch_run/control.rcc",
                        },
                        "samples": [
                            {
                                "cimac_id": "CTTTPP101.00",
                                "raw_rcc": {
                                    "object_url": f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/nanostring_batch_run/CTTTPP101.00.rcc",
                                },
                            },
                        ],
                    },
                    {
                        "run_id": "nanostring_batch_run_2",
                        "control_raw_rcc": {
                            "object_url": f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/nanostring_batch_run_2/control.rcc",
                        },
                        "samples": [
                            {
                                "cimac_id": "CTTTPP201.00",
                                "raw_rcc": {
                                    "object_url": f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/nanostring_batch_run_2/CTTTPP201.00.rcc",
                                },
                            },
                        ],
                    },
                ],
            },
            {
                "batch_id": "nanostring_batch_2",
                "data": {
                    "normalized": {
                        "object_url": f"{TEST_TRIAL_ID}/nanostring/nanostring_batch_2/normalized_data.csv",
                    },
                },
                "runs": [
                    {
                        "run_id": "nanostring_batch_2_run",
                        "control_raw_rcc": {
                            "object_url": f"{TEST_TRIAL_ID}/nanostring/nanostring_batch_2/nanostring_batch_2_run/control.rcc",
                        },
                        "samples": [
                            {
                                "cimac_id": "CTTTPP102.00",
                                "raw_rcc": {
                                    "object_url": f"{TEST_TRIAL_ID}/nanostring/nanostring_batch_2/nanostring_batch_2_run/CTTTPP102.00.rcc",
                                },
                            },
                        ],
                    },
                ],
            },
        ],
        "wes": [
            {
                "records": [
                    {
                        "cimac_id": "CTTTPP101.00",
                        "files": {
                            "bam": [
                                {
                                    "object_url": f"{TEST_TRIAL_ID}/wes/CTTTPP101.00/reads_0.bam",
                                },
                                {
                                    "object_url": f"{TEST_TRIAL_ID}/wes/CTTTPP101.00/reads_1.bam",
                                },
                            ],
                        },
                    },
                    {
                        "cimac_id": "CTTTPP201.00",
                        "files": {
                            "bam": [
                                {
                                    "object_url": f"{TEST_TRIAL_ID}/wes/CTTTPP201.00/reads_0.bam",
                                },
                            ],
                        },
                    },
                ],
            },
            {
                "records": [
                    {
                        "cimac_id": "CTTTPP102.00",
                        "files": {
                            "bam": [
                                {
                                    "object_url": f"{TEST_TRIAL_ID}/wes/CTTTPP102.00/reads_0.bam",
                                },
                            ],
                        },
                    },
                ],
            },
        ],
    },
}



import os
from pathlib import Path
import click

from . import cache

# General config
CIDC_WORKING_DIR = os.path.join(Path.home(), ".cidc")
TOKEN_CACHE_PATH = os.path.join(CIDC_WORKING_DIR, "id_token")


# Environment management
_ENV_KEY = "env"


def set_env(value: str) -> None:
    """Set the current CLI environment"""
    cache.store(_ENV_KEY, value)


def get_env() -> str:
    """Get the current CLI environment"""
    return cache.get(_ENV_KEY) or "prod"


_WARNING = "\n".join(
    [
        "##      ##    ###    ########  ##    ## #### ##    ##  ######   ",
        "##  ##  ##   ## ##   ##     ## ###   ##  ##  ###   ## ##    ##  ",
        "##  ##  ##  ##   ##  ##     ## ####  ##  ##  ####  ## ##        ",
        "##  ##  ## ##     ## ########  ## ## ##  ##  ## ## ## ##   #### ",
        "##  ##  ## ######### ##   ##   ##  ####  ##  ##  #### ##    ##  ",
        "##  ##  ## ##     ## ##    ##  ##   ###  ##  ##   ### ##    ##  ",
        " ###  ###  ##     ## ##     ## ##    ## #### ##    ##  ######   ",
        "",
    ]
)
_STRIKE = "*" * 64


def check_env_warning(ignore_env) -> None:
    """Get the current CLI environment"""
    if get_env() != "prod":
        print(_STRIKE + "\n" + _WARNING + _STRIKE)
        print(f"You are using DEVELOPMENT environment ({get_env()})")
        if ignore_env != None and ignore_env == get_env():
            return

        print("If you are not sure what that means, stop now.\n" + _STRIKE)

        if not click.confirm("Proceed anyways?"):
            if click.confirm(
                "Do you want to reset the CLI to its default configuration?"
            ):
                set_env("prod")
                print(_STRIKE)
                print("Environment set to default, you can retry now.")
            exit(0)

        print(_STRIKE)

    if ignore_env != None and ignore_env != get_env():
        print(_STRIKE + "\n" + _WARNING + _STRIKE)
        print(f"You are using PRODUCTION environment, not {ignore_env}")
        print(f"Remove `--ignore {ignore_env}` and retry.")
        print(_STRIKE)
        exit(0)


# Environment-specific config
_current_env = get_env()
if _current_env == "prod":
    API_V2_URL = "https://nih-nci-cidc-datapipelines.uk.r.appspot.com"
elif _current_env == "staging":
    API_V2_URL = "https://nih-nci-cimac-cidc-dpline-stg.uk.r.appspot.com"
elif _current_env == "stage-cloudtwo":
    API_V2_URL = "https://nih-nci-cimac-cidc-stage.uk.r.appspot.com"
elif _current_env == "dev-int":
    API_V2_URL = "https://nih-nci-cimac-cidc-dpline-dev.uk.r.appspot.com"
elif _current_env == "dev-cloudtwo":
    API_V2_URL = "https://nih-nci-cimac-cidc-dev2.uk.r.appspot.com"
elif _current_env == "dev":
    API_V2_URL = "http://localhost:8000"
else:
    raise ValueError(f"Unsupported environment: {_current_env}")



import getpass
from types import ModuleType
from typing import List, Optional
import warnings
from .config import get_username, set_username

from google.cloud.sql.connector import Connector
import sqlalchemy
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import sessionmaker

from ..config import get_env

global Session
global DownloadableFiles, TrialMetadata, UploadJobs, Users

Session = None
DownloadableFiles, TrialMetadata, UploadJobs, Users = None, None, None, None


def connect(list_mod: ModuleType) -> None:
    """
    Set up this module to be able to make sqlalchemy calls
    uses ENV as set by $ nci-cidc config set-env
    defaults to staging unless `prod` is specified ie no `dev` mode

    If not already loaded, will ask for your database username
    Asks for database password every time as to not store it
    """
    ENV: str = get_env()
    # TODO: support dev environment running database locally
    if ENV not in ["prod", "staging", "dev-int", "dev-cloudtwo", "stage=cloudtwo"]:
        print("unknown ENV, applying to dev-int:", ENV)
        ENV: str = "dev-int"

    username: Optional[str] = get_username()
    if not username:
        username = input("Username: ")
        set_username(username)

    password = getpass.getpass()
    connections = {
        "prod": "nih-nci-cidc-datapipelines:us-east4:cidc-postgresql-cidc-prod2",
        "staging": "nih-nci-cimac-cidc-dpline-stg:us-east4:cidc-postgresql-cidc-stage2",
        "dev-int": "nih-nci-cimac-cidc-dpline-dev:us-east4:cidc-postgresql-cidc-dev2",
        "dev-cloudtwo": "nih-nci-cimac-cidc-dev2:us-east4:cidc-postgresql-dev2",
        "stage-cloudtwo": "nih-nci-cimac-cidc-stage:us-east4:cidc-postgresql-stage2"
    }
    connection_name = connections[ENV]

    # initialize Connector object
    connector = Connector()

    # function to return the database connection
    def getconn():
        databases = {
            "prod": "cidc-cidc-prod2",
            "staging": "cidc-cidc-stage2",
            "dev-int": "cidc-cidc-dev2",
            "dev-cloudtwo": "cidc-cidc-dev2",
            "stage-cloudtwo": "cidc-cidc-stage2"
        }
        conn = connector.connect(
            connection_name,
            "pg8000",
            user=username,
            password=password,
            db=databases[ENV],
        )
        return conn

    Base = automap_base()
    engine = sqlalchemy.create_engine("postgresql+pg8000://", creator=getconn)

    global Session
    Session = sessionmaker(engine)

    # from here, Base.classes will have each table as an attribute
    # they are sqlalchemy tables equivalent to the ones in API/models/models.py
    # eg Base.classes.trial_metadata works like cidc_api.models.TrialMetadata
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=sqlalchemy.exc.SAWarning)
        Base.prepare(
            engine,
            reflection_options={
                "only": ["downloadable_files", "trial_metadata", "upload_jobs", "users"]
            },
        )

    # these won't have tab complete because they're reflecting not defining
    global DownloadableFiles, TrialMetadata, UploadJobs, Users

    DownloadableFiles = Base.classes.downloadable_files
    TrialMetadata = Base.classes.trial_metadata
    UploadJobs = Base.classes.upload_jobs
    Users = Base.classes.users

    # wire up the passed module
    # this is a hack to not pass them around
    list_mod.Session = Session
    list_mod.DownloadableFiles = DownloadableFiles
    list_mod.TrialMetadata = TrialMetadata
    list_mod.UploadJobs = UploadJobs
    list_mod.Users = Users


def get_clinical_downloadable_files(
    trial_id: str, *, session: Session
) -> List[DownloadableFiles]:
    """
    Get all downloadable_files rows from clinical data files for the given trial

    Parameters
    ----------
    trial_id: str
        the id of the trial to affect
    session: Session
        a session created from this module's `Session` after `connect()` is called
    """
    return (
        session.query(DownloadableFiles)
        .filter(
            DownloadableFiles.trial_id == trial_id,
            DownloadableFiles.object_url.like("%/clinical/%"),
        )
        .all()
    )


def get_misc_data_files(trial_id: str, *, session: Session) -> List[DownloadableFiles]:
    """
    Get all downloadable_files rows from misc_data uploads for the given trial

    Parameters
    ----------
    trial_id: str
        the id of the trial to affect
    session: Session
        a session created from this module's `Session` after `connect()` is called
    """
    return (
        session.query(DownloadableFiles)
        .filter(
            DownloadableFiles.trial_id == trial_id,
            DownloadableFiles.object_url.like("%/misc_data/%"),
        )
        .all()
    )


def get_shipments(trial_id: str, *, session: Session) -> List[UploadJobs]:
    """
    Get all upload_jobs rows from successful manifest uploads for the given trial

    Parameters
    ----------
    trial_id: str
        the id of the trial to affect
    session: Session
        a session created from this module's `Session` after `connect()` is called
    """
    # the "shipments" in trial_metadata don't have which
    # samples they came with, so we have to check the uploads
    shipments = []
    for u in (
        session.query(UploadJobs)
        .filter(
            UploadJobs.trial_id == trial_id,
            UploadJobs.status == "merge-completed",
        )
        .all()
    ):
        # only for shipment uploads
        if "shipments" in u.metadata_patch:
            shipments.append(u)
    return shipments


def get_trial_if_exists(
    trial_id: str, *, with_for_update: bool = False, session: Session
) -> TrialMetadata:
    """
    Get the trial_metadata row for the given trial for update
    Exits with message if trial does not exist

    Parameters
    ----------
    trial_id: str
        the id of the trial to affect
    session: Session
        a session created from this module's `Session` after `connect()` is called
    """
    query = session.query(TrialMetadata).filter(TrialMetadata.trial_id == trial_id)
    if with_for_update:
        query = query.with_for_update()
    trial: TrialMetadata = query.first()

    if not trial:
        print(f"Trial {trial_id} cannot be found")
        exit()

    return trial



import sys
from io import BytesIO, StringIO

import click
import pytest
from unittest.mock import MagicMock
from typing import Union

from nci_cidc_cli import api, auth, config, cache, __version__


def make_json_response(body={}) -> MagicMock:
    response = MagicMock()
    response.json.return_value = body
    response.status_code = 200
    return response


def make_error_response(msg: Union[str, list], code: int = 400) -> MagicMock:
    response = make_json_response({"_error": {"message": msg}})
    response.status_code = code
    return response


def patch_request(http_verb, response, monkeypatch):
    monkeypatch.setattr(f"requests.{http_verb}", lambda *args, **kwargs: response)


def test_url_builder():
    base_url = config.API_V2_URL + "/"
    assert api._url("") == base_url
    assert api._url("/a") == base_url + "a"
    assert api._url("a") == base_url + "a"


def test_error_message_extractor():
    MSG = "an error message"
    response = make_error_response(MSG)
    assert api._error_message(response) == MSG

    # Error response without message
    response = MagicMock()
    response.json.side_effect = Exception()
    response.status_code = 503
    assert "API server encountered an error" in api._error_message(response)

    MSG = ["first", "another"]
    response = make_error_response(MSG)
    for m in MSG:
        assert m in api._error_message(response)

    # Error response without proper json _error message
    response = MagicMock()
    response.json.side_effect = Exception()
    response.status_code = 503
    assert "API server encountered an error" in api._error_message(response)

    # Error response without proper json _error and 4xx code
    response = MagicMock()
    response.json.side_effect = Exception()
    response.status_code = 403
    assert "API server encountered an error" in api._error_message(response)


def test_with_auth(monkeypatch):
    """Test the authorization header builder"""
    TOKEN = "tok"
    AUTH_HEADER = {
        "Authorization": f"Bearer {TOKEN}",
        "User-Agent": f"nci-cidc-cli/{__version__}",
    }
    OTHER_HEADERS = {"If-Match": "blah blah blah", "Content-Type": "application/json"}
    HEADERS = {**OTHER_HEADERS, **AUTH_HEADER}

    assert api._with_auth(id_token=TOKEN) == AUTH_HEADER
    assert api._with_auth(headers=OTHER_HEADERS, id_token=TOKEN) == HEADERS

    monkeypatch.setattr("nci_cidc_cli.auth.validate_token", MagicMock())
    auth.validate_and_cache_token(TOKEN)
    assert api._with_auth() == AUTH_HEADER
    assert api._with_auth(headers=OTHER_HEADERS) == HEADERS


def test_check_auth(monkeypatch):
    """Check that api.check_auth handles errors as expected"""
    # Auth error
    ERR = "signature expired"
    not_authorized = make_error_response(ERR, 401)
    patch_request("get", not_authorized, monkeypatch)
    with pytest.raises(api.ApiError, match=ERR):
        api.check_auth("foo_token")

    # Successful authorization
    success = make_json_response()
    patch_request("get", success, monkeypatch)
    assert api.check_auth("foo_token") is None


def test_list_assays(monkeypatch):
    assays = ["wes", "pbmc"]
    response = make_json_response(assays)
    patch_request("get", response, monkeypatch)
    assert api.list_assays() == assays


JOB_ID = 1
JOB_ETAG = "abcd"
UPLOAD_TOKEN = "test-upload-token"
UPLOAD_URL_MAP = {}


def test_initiate_upload(monkeypatch):
    """Test upload initation builds a request and parses a response correctly"""
    ASSAY = "wes"
    XLSX = BytesIO(b"abcd")
    GCS_BUCKET = "bucket"
    URL_MAPPING = {"foo": "bar"}
    EXTRA_METADATA = {"uuid": "lp1"}
    GCS_FILE_MAP = {"bar": "baz"}
    OPTIONAL_FILES = []

    monkeypatch.setattr(api, "_with_auth", lambda: {})

    def good_request(url, headers, data, files):
        assert data.get("schema") == "wes"
        assert files.get("template") == XLSX
        return make_json_response(
            {
                "job_id": JOB_ID,
                "job_etag": JOB_ETAG,
                "gcs_bucket": GCS_BUCKET,
                "url_mapping": URL_MAPPING,
                "extra_metadata": EXTRA_METADATA,
                "gcs_file_map": GCS_FILE_MAP,
                "optional_files": OPTIONAL_FILES,
                "token": UPLOAD_TOKEN,
            }
        )

    monkeypatch.setattr("requests.post", good_request)
    api.initiate_upload(ASSAY, XLSX)

    ERR = "bad request or something"
    bad_request = make_error_response(ERR, 400)
    patch_request("post", bad_request, monkeypatch)
    with pytest.raises(api.ApiError, match=ERR):
        api.initiate_upload(ASSAY, XLSX)

    BADLY_TYPED_ERR = [ERR]
    bad_request = make_error_response(BADLY_TYPED_ERR, 400)
    patch_request("post", bad_request, monkeypatch)
    with pytest.raises(api.ApiError, match=ERR):
        api.initiate_upload(ASSAY, XLSX)

    cant_decode = make_json_response({"foo": "bar"})
    patch_request("post", cant_decode, monkeypatch)
    with pytest.raises(api.ApiError, match="Cannot decode API response"):
        api.initiate_upload(ASSAY, XLSX)


def test_update_job_status(monkeypatch):
    """Test that _update_job_status builds a request with the expected structure"""
    monkeypatch.setattr(api, "_with_auth", lambda headers: headers)

    def test_status(status):
        def request(url, params, json, headers):
            assert url.endswith("/upload_jobs/" + str(JOB_ID))
            assert json == {"status": status, "gcs_file_map": {}}
            assert headers.get("If-Match") == JOB_ETAG
            assert params["token"] == UPLOAD_TOKEN
            return make_json_response()

        return request

    monkeypatch.setattr("requests.patch", test_status("upload-completed"))
    api.upload_succeeded(JOB_ID, UPLOAD_TOKEN, JOB_ETAG, UPLOAD_URL_MAP)

    monkeypatch.setattr("requests.patch", test_status("upload-failed"))
    api.upload_failed(JOB_ID, UPLOAD_TOKEN, JOB_ETAG, UPLOAD_URL_MAP)


def test_poll_upload_merge_status(monkeypatch):
    """Check that poll_upload_merge status handles various responses as expected"""
    monkeypatch.setattr(api, "_with_auth", lambda: {})

    def not_found_get(*args, **kwargs):
        return make_error_response("", code=404)

    monkeypatch.setattr("requests.get", not_found_get)
    with pytest.raises(api.ApiError):
        api.poll_upload_merge_status(1, UPLOAD_TOKEN)

    def bad_response_get(*args, **kwargs):
        return make_json_response({})

    monkeypatch.setattr("requests.get", bad_response_get)
    with pytest.raises(api.ApiError, match="unexpected upload status message"):
        api.poll_upload_merge_status(1, UPLOAD_TOKEN)

    def good_retry_get(*args, **kwargs):
        return make_json_response({"retry_in": 5})

    monkeypatch.setattr("requests.get", good_retry_get)
    upload_status = api.poll_upload_merge_status(1, UPLOAD_TOKEN)
    assert upload_status.retry_in == 5
    assert upload_status.status is None
    assert upload_status.status_details is None

    status_res = {"status": "merge-failed", "status_details": "error message here"}

    def good_status_get(*args, **kwargs):
        return make_json_response(status_res)

    monkeypatch.setattr("requests.get", good_status_get)
    upload_status = api.poll_upload_merge_status(1, UPLOAD_TOKEN)
    assert upload_status.retry_in is None
    assert upload_status.status == status_res["status"]
    assert upload_status.status_details == status_res["status_details"]


def test_retry_with_reauth(runner, capsys, monkeypatch):
    """Ensure retry-after-reauthentication decorator works as expected."""

    @api.retry_with_reauth
    def req_200():
        return make_json_response({"foo": "bar"})

    # No interference with a successful request
    res = req_200()
    assert res.json() == {"foo": "bar"}

    @api.retry_with_reauth
    def req_500():
        return make_error_response("uhoh", code=500)

    # Raises ApiError on request that fails for non-auth reasons
    with pytest.raises(api.ApiError, match="uhoh"):
        res = req_500()

    @api.retry_with_reauth
    def req_401_perms():
        return make_error_response(
            {
                "errors": [
                    "test@user.com is not authorized to upload pbmc data to some trial."
                ]
            }
        )

    # Raises ApiError on request that fails for permissions reasons
    with pytest.raises(api.ApiError, match="is not authorized to upload pbmc data"):
        res = req_401_perms()

    good_token = "good_token"

    @api.retry_with_reauth
    def req_401(*args, **kwargs):
        if "files" in kwargs:
            # Ensure that files have been rewinded before retrying a request
            for file in kwargs["files"].values():
                assert len(file.read()) > 0
        try:
            token = auth.get_id_token()
            if token == good_token:
                return make_json_response("successful reauth")
        except:
            pass
        return make_error_response("auth error", code=401)

    with runner.isolated_filesystem():
        bad_token = "bad_token"
        # Bypass auth.validate_and_cache_token's validation functionality
        cache.store(auth.TOKEN, bad_token)

        def successful_reauth(*args):
            pass

        # Simulate a user entering a fresh, valid token
        monkeypatch.setattr(api, "_read_clipboard", lambda: good_token)
        monkeypatch.setattr(api, "check_auth", successful_reauth)

        monkeypatch.setattr("sys.stdin", StringIO("\n"))
        res = req_401()
        stdout = capsys.readouterr().out
        assert res.json() == "successful reauth"
        # User is prompted once, and their token is displayed once
        assert stdout.count("paste your copied token below") == 1
        assert stdout.count(good_token) == 1

        # Test that files are rewound before requests are retried
        monkeypatch.setattr("sys.stdin", StringIO("\n"))
        cache.store(auth.TOKEN, bad_token)
        files = {"a": BytesIO(b"a"), "b": BytesIO(b"b")}
        res = req_401(files=files)
        assert files["a"].read() == b""
        assert files["b"].read() == b""
        assert res.json() == "successful reauth"

        def unsuccessful_reauth(*args):
            raise api.ApiError("signature expired")

        # Simulate a user entering 3 invalid tokens
        cache.store(auth.TOKEN, bad_token)
        monkeypatch.setattr(api, "_read_clipboard", lambda: bad_token)
        monkeypatch.setattr("sys.stdin", StringIO("\n" * 3))
        monkeypatch.setattr(api, "check_auth", unsuccessful_reauth)

        # This will be raised when mocked stdin runs out
        with pytest.raises(click.exceptions.Abort):
            req_401()

        stdout = capsys.readouterr().out
        # User enters bad_token 3 times
        assert stdout.count(bad_token) == 3
        # User is re-prompted 5 times
        assert stdout.count("paste your copied token below") == 5

        # Simulate a user having pyperclip issues
        def throw():
            raise Exception

        monkeypatch.setattr(api, "_read_clipboard", throw)
        monkeypatch.setattr("sys.stdin", StringIO("\n" * 2))
        auth.cache.store(auth.TOKEN, bad_token)

        with pytest.raises(api.ApiError, match="auth error"):
            req_401()

        stdout = capsys.readouterr().out
        assert stdout.count("could not read token from clipboard") == 1



import pytest

from nci_cidc_cli import auth


def test_get_user_email(monkeypatch):
    """Check that get_user_email extracts an email from a JWT"""
    # This is a JWT with payload {"email": "test@email.com"}
    TOKEN = "eyJhbGciOiJIUzI1NiJ9.eyJlbWFpbCI6InRlc3RAZW1haWwuY29tIn0.jTck01ns477JzZY-KLMM82OOZcc6ctJA11HKE8sXiq4"

    monkeypatch.setattr(auth, "get_id_token", lambda: TOKEN)

    email = auth.get_user_email()

    assert email == "test@email.com"

    # Test that it handles invalid JWTs
    monkeypatch.setattr(auth, "get_id_token", lambda: "uh oh")
    with pytest.raises(auth.AuthError, match="not authenticated"):
        auth.get_user_email()


def test_valid_token_flow(monkeypatch, runner):
    """Check that caching works as expected for a valid token"""
    monkeypatch.setattr(auth, "validate_token", lambda token: None)

    TOKEN = "test-token"

    with runner.isolated_filesystem():
        # Login
        auth.validate_and_cache_token(TOKEN)

        # Use the token
        assert auth.get_id_token() == TOKEN


def test_invalid_token_flow(monkeypatch, runner):
    """Check that errors are thrown as expected for an invalid token"""

    def auth_error(*args):
        raise auth.AuthError("uh oh")

    monkeypatch.setattr(auth, "validate_token", auth_error)

    with runner.isolated_filesystem():
        # Invalid tokens shouldn't be cached
        with pytest.raises(auth.AuthError):
            auth.validate_and_cache_token("blah")

        # Invalid tokens *can* be read.
        monkeypatch.setattr("nci_cidc_cli.cache.get", lambda key: "blah")
        assert auth.get_id_token() == "blah"



"""Utilities for working with a user's gcloud installation"""
import shutil
import subprocess

import click

from . import auth

GCLOUD = "gcloud"


def check_installed():
    """Check if gcloud appears to be installed on this computer."""
    if not shutil.which(GCLOUD):
        raise click.ClickException(
            "The CIDC CLI requires an installation of the gcloud SDK. "
            "To install, please visit: https://cloud.google.com/sdk/install"
        )


def login():
    """Check if a user is logged in to gcloud, and log them in if not."""
    email = auth.get_user_email()

    # Try to log the user in to gcloud with their CIDC email
    click.secho("$ gcloud auth login --no-launch-browser --brief", dim=True)
    subprocess.call([GCLOUD, "auth", "login", email, "--no-launch-browser", "--brief"])



from copy import deepcopy
from datetime import datetime
from deepdiff import DeepDiff
import pytest
from _pytest.monkeypatch import MonkeyPatch
from typing import Dict, List
from unittest.mock import MagicMock, call

from nci_cidc_cli.dbedit import remove as dbedit_remove
from .constants import (
    TEST_CLINICAL_URL_CSV,
    TEST_CLINICAL_URL_XLSX,
    TEST_MANIFEST_ID,
    TEST_METADATA_JSON,
    TEST_TRIAL_ID,
)

CLIPPED_METADATA_SHIPMENTS: dict = {
    "analysis": TEST_METADATA_JSON["analysis"],
    "assays": TEST_METADATA_JSON["assays"],
    "clinical_data": TEST_METADATA_JSON["clinical_data"],
    "participants": [
        {
            "cimac_participant_id": "CTTTPP1",
            "samples": [
                {"cimac_id": "CTTTPP102.00"},
            ],
        },
        {
            "cimac_participant_id": "CTTTPP3",
            "samples": [
                {"cimac_id": "CTTTPP301.00"},
                {"cimac_id": "CTTTPP303.00"},
            ],
        },
    ],
    "shipments": [
        {"manifest_id": TEST_MANIFEST_ID + "2"},
    ],
}
CLIPPED_METADATA_TARGET_CLINICAL: dict = {
    "analysis": TEST_METADATA_JSON["analysis"],
    "assays": TEST_METADATA_JSON["assays"],
    "clinical_data": {
        "records": [
            {
                "clinical_file": {
                    "object_url": f"{TEST_TRIAL_ID}/clinical/{TEST_CLINICAL_URL_CSV}",
                    "number_of_participants": 3,
                },
            },
        ],
    },
    "participants": TEST_METADATA_JSON["participants"],
    "shipments": TEST_METADATA_JSON["shipments"],
}
CLIPPED_METADATA_ALL_CLINICAL: dict = {
    "analysis": TEST_METADATA_JSON["analysis"],
    "assays": TEST_METADATA_JSON["assays"],
    "participants": TEST_METADATA_JSON["participants"],
    "shipments": TEST_METADATA_JSON["shipments"],
}


def test_remove_clinical(monkeypatch):
    # mock the class
    DownloadableFiles = MagicMock()
    monkeypatch.setattr(dbedit_remove, "DownloadableFiles", DownloadableFiles)

    # mock the session itself
    Session = MagicMock()
    session = MagicMock()
    begin = MagicMock()

    # mock downlodable files
    mock_files = [MagicMock(), MagicMock()]
    mock_files[0].object_url = f"{TEST_TRIAL_ID}/clinical/{TEST_CLINICAL_URL_XLSX}"
    mock_files[1].object_url = f"{TEST_TRIAL_ID}/clinical/{TEST_CLINICAL_URL_CSV}"

    # mock query for individual file
    query = MagicMock()
    filter_query = MagicMock()
    filter_query.all.return_value = [mock_files[0]]
    query.filter.return_value = filter_query

    session.query.return_value = query
    begin.__enter__.return_value = session
    Session.begin.return_value = begin
    monkeypatch.setattr(dbedit_remove, "Session", Session)

    # mock getting all clinical files
    get_clinical_downloadable_files = MagicMock()
    get_clinical_downloadable_files.return_value = mock_files
    monkeypatch.setattr(
        dbedit_remove,
        "get_clinical_downloadable_files",
        get_clinical_downloadable_files,
    )

    # mock getting the trial
    get_trial_if_exists = MagicMock()
    mock_trial = MagicMock()
    mock_trial.metadata_json = deepcopy(TEST_METADATA_JSON)
    mock_trial._updated = datetime.fromisoformat("2020-01-01T12:34:45")
    get_trial_if_exists.return_value = mock_trial
    monkeypatch.setattr(dbedit_remove, "get_trial_if_exists", get_trial_if_exists)

    # mock table class
    TrialMetadata = MagicMock()
    monkeypatch.setattr(dbedit_remove, "TrialMetadata", TrialMetadata)

    # convenience function
    def reset_mocks():
        Session.reset_mock()
        session.reset_mock()
        begin.reset_mock()
        query.reset_mock()
        filter_query.reset_mock()
        get_clinical_downloadable_files.reset_mock()
        [mock.reset_mock() for mock in mock_files]
        get_trial_if_exists.reset_mock()
        mock_trial.reset_mock()

    # test for proper removal of single file
    reset_mocks()
    dbedit_remove.remove_clinical(TEST_TRIAL_ID, TEST_CLINICAL_URL_XLSX)

    Session.begin.assert_called_once_with()
    begin.__enter__.assert_called_once()
    begin.__exit__.assert_called_once()
    get_trial_if_exists.assert_called_once_with(
        TEST_TRIAL_ID, with_for_update=True, session=session
    )
    get_clinical_downloadable_files.assert_not_called()
    assert all(
        test_call in query.filter.call_args_list
        for test_call in [
            call(
                DownloadableFiles.trial_id == TEST_TRIAL_ID,
                DownloadableFiles.object_url
                == f"{TEST_TRIAL_ID}/clinical/{TEST_CLINICAL_URL_XLSX}",
            ),
            call(
                TrialMetadata.trial_id == TEST_TRIAL_ID,
            ),
        ]
    )
    filter_query.all.assert_called_once_with()
    filter_query.update.assert_called_once()
    args, _ = filter_query.update.call_args
    assert len(args) == 1

    assert isinstance(args[0][TrialMetadata._updated], datetime)
    assert args[0][TrialMetadata._updated] != datetime.fromisoformat(
        "2020-01-01T12:34:45"
    )
    assert (
        DeepDiff(args[0][TrialMetadata.metadata_json], CLIPPED_METADATA_TARGET_CLINICAL)
        == {}
    )

    session.delete.assert_called_once_with(mock_files[0])

    # test for proper removal of all files
    reset_mocks()
    dbedit_remove.remove_clinical(TEST_TRIAL_ID, "*")

    Session.begin.assert_called_once_with()
    begin.__enter__.assert_called_once()
    begin.__exit__.assert_called_once()
    get_trial_if_exists.assert_called_once_with(
        TEST_TRIAL_ID,
        with_for_update=True,
        session=session,
    )
    get_clinical_downloadable_files.assert_called_once_with(
        TEST_TRIAL_ID,
        session=session,
    )

    filter_query.update.assert_called_once()
    args, _ = filter_query.update.call_args
    assert len(args) == 1
    assert isinstance(args[0][TrialMetadata._updated], datetime)

    assert args[0][TrialMetadata._updated] != datetime.fromisoformat(
        "2020-01-01T12:34:45"
    )
    assert (
        DeepDiff(args[0][TrialMetadata.metadata_json], CLIPPED_METADATA_ALL_CLINICAL)
        == {}
    )

    assert session.delete.call_count == 2
    session.delete.assert_has_calls(
        [
            call(mock_files[0]),
            call(mock_files[1]),
        ]
    )

    # check that it exits if no matching file found
    reset_mocks()
    filter_query.all.return_value = []
    query.filter.return_value = filter_query
    session.query.return_value = query
    begin.__enter__.return_value = session
    Session.begin.return_value = begin
    monkeypatch.setattr(dbedit_remove, "Session", Session)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        dbedit_remove.remove_clinical(
            trial_id=TEST_TRIAL_ID,
            target_id=TEST_CLINICAL_URL_XLSX,
        )
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
    session.add.assert_not_called()
    session.delete.assert_not_called()

    # check that it exits if no files found for *
    reset_mocks()
    get_clinical_downloadable_files.return_value = []
    monkeypatch.setattr(
        dbedit_remove,
        "get_clinical_downloadable_files",
        get_clinical_downloadable_files,
    )

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        dbedit_remove.remove_clinical(
            trial_id=TEST_TRIAL_ID,
            target_id="*",
        )
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
    session.add.assert_not_called()
    session.delete.assert_not_called()


def test_remove_samples_from_blob():
    samples_to_remove: Dict[str, List[str]] = {
        "CTTTPP1": ["CTTTPP101.00"],
        "CTTTPP2": ["CTTTPP201.00", "CTTTPP202.00"],
        "CTTTPP3": ["CTTTPP302.00"],
    }

    res = dbedit_remove._remove_samples_from_blob(
        metadata_json=TEST_METADATA_JSON,
        samples_to_remove=samples_to_remove,
    )

    # this doesn't remove the shipment itself
    clip_just_samples = deepcopy(CLIPPED_METADATA_SHIPMENTS)
    clip_just_samples["shipments"] = TEST_METADATA_JSON["shipments"]
    assert res == clip_just_samples


def test_remove_shipment(monkeypatch):
    # mock the session itself
    Session = MagicMock()
    session = MagicMock()
    begin = MagicMock()
    query = MagicMock()
    filter_query = MagicMock()

    query.filter.return_value = filter_query
    session.query.return_value = query
    begin.__enter__.return_value = session
    Session.begin.return_value = begin
    monkeypatch.setattr(dbedit_remove, "Session", Session)

    # mock getting the trial
    get_trial_if_exists = MagicMock()
    mock_trial = MagicMock()
    mock_trial.metadata_json = TEST_METADATA_JSON
    mock_trial._updated = datetime.fromisoformat("2020-01-01T12:34:45")
    get_trial_if_exists.first.return_value = mock_trial
    monkeypatch.setattr(dbedit_remove, "get_trial_if_exists", get_trial_if_exists)

    # mock getting the uploads
    mock_uploads = [MagicMock(), MagicMock(), MagicMock()]
    mock_uploads[0].metadata_patch = {
        "participants": [
            {
                "cimac_participant_id": "CTTTPP1",
                "samples": [
                    {"cimac_id": "CTTTPP101.00"},
                ],
            },
        ],
        "shipments": [{"manifest_id": TEST_MANIFEST_ID}],
    }  # TARGET: should be returned
    mock_uploads[1].metadata_patch = {
        "participants": [
            {
                "cimac_participant_id": "CTTTPP1",
                "samples": [
                    {"cimac_id": "CTTTPP102.00"},
                ],
            },
            {
                "cimac_participant_id": "CTTTPP3",
                "samples": [
                    {"cimac_id": "CTTTPP302.00"},
                ],
            },
        ],
        "shipments": [{"manifest_id": TEST_MANIFEST_ID + "2"}],
    }  # should be returned

    mock_uploads[2].metadata_patch = {
        "participants": [
            {
                "cimac_participant_id": "CTTTPP2",
                "samples": [
                    {"cimac_id": "CTTTPP201.00"},
                    {"cimac_id": "CTTTPP202.00"},
                ],
            },
            {
                "cimac_participant_id": "CTTTPP3",
                "samples": [
                    {"cimac_id": "CTTTPP301.00"},
                    {"cimac_id": "CTTTPP303.00"},
                ],
            },
        ],
        "shipments": [{"manifest_id": TEST_MANIFEST_ID}],
    }
    get_shipments = MagicMock()
    get_shipments.return_value = mock_uploads
    monkeypatch.setattr(dbedit_remove, "get_shipments", get_shipments)

    # mock table class
    TrialMetadata = MagicMock()
    monkeypatch.setattr(dbedit_remove, "TrialMetadata", TrialMetadata)

    # convenience function
    def reset_mocks():
        Session.reset_mock()
        session.reset_mock()
        begin.reset_mock()
        get_trial_if_exists.reset_mock()
        mock_trial.reset_mock()
        [mock.reset_mock() for mock in mock_uploads]
        get_shipments.reset_mock()

    # test for proper removal
    reset_mocks()
    dbedit_remove.remove_shipment(
        trial_id=TEST_TRIAL_ID,
        target_id=TEST_MANIFEST_ID,
    )

    Session.begin.assert_called_once_with()
    begin.__enter__.assert_called_once()
    begin.__exit__.assert_called_once()
    get_trial_if_exists.assert_called_once_with(
        TEST_TRIAL_ID, with_for_update=True, session=session
    )
    get_shipments.assert_called_once_with(TEST_TRIAL_ID, session=session)

    args, _ = filter_query.update.call_args
    assert len(args) == 1

    assert isinstance(args[0][TrialMetadata._updated], datetime)
    assert args[0][TrialMetadata._updated] != datetime.fromisoformat(
        "2020-01-01T12:34:45"
    )
    args[0][TrialMetadata.metadata_json] == CLIPPED_METADATA_SHIPMENTS

    assert session.delete.call_count == 2
    session.delete.assert_has_calls(
        [
            call(mock_uploads[0]),
            call(mock_uploads[2]),
        ]
    )

    # check that it exits if no samples are in the matching manifest
    reset_mocks()
    remove_samples_from_blob = MagicMock()
    monkeypatch.setattr(
        dbedit_remove, "_remove_samples_from_blob", remove_samples_from_blob
    )

    empty_shipment = MagicMock()
    empty_shipment.metadata_json = {
        "participants": [],
        "shipments": [
            {"manifest_id": TEST_MANIFEST_ID},
        ],
    }
    get_shipments.return_value = [empty_shipment]
    monkeypatch.setattr(dbedit_remove, "get_shipments", get_shipments)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        dbedit_remove.remove_shipment(
            trial_id=TEST_TRIAL_ID,
            target_id=TEST_MANIFEST_ID,
        )
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
    remove_samples_from_blob.assert_not_called()
    session.add.assert_not_called()
    session.delete.assert_not_called()

    # check that it exits if no matching manifest found
    reset_mocks()
    remove_samples_from_blob.reset_mock()
    get_shipments.return_value = []
    monkeypatch.setattr(dbedit_remove, "get_shipments", get_shipments)

    with pytest.raises(SystemExit) as pytest_wrapped_e:
        dbedit_remove.remove_shipment(
            trial_id=TEST_TRIAL_ID,
            target_id=TEST_MANIFEST_ID,
        )
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
    remove_samples_from_blob.assert_not_called()
    session.add.assert_not_called()
    session.delete.assert_not_called()


class Test_remove_data:
    def teardown(self):
        self.monkeypatch.setattr("builtins.print", self.real_print)

    def setup(self):
        self.Session = MagicMock()
        self.begin = MagicMock()
        self.session = MagicMock()
        self.query = MagicMock()
        self.filter_query = MagicMock()

        self.query.filter.return_value = self.filter_query
        self.session.query.return_value = self.query
        self.begin.__enter__.return_value = self.session
        self.Session.begin.return_value = self.begin

        self.mock_trial = MagicMock()
        self.mock_trial.metadata_json = deepcopy(TEST_METADATA_JSON)

        self.get_trial_if_exists = MagicMock()
        self.get_trial_if_exists.return_value = self.mock_trial

        self.mock_remove_clinical = MagicMock()
        self.mock_remove_misc_data = MagicMock()

        self.real_print = print
        self.mock_print = MagicMock()

        self.DownloadableFiles = MagicMock()
        self.TrialMetadata = MagicMock()

        self.monkeypatch = MonkeyPatch()
        self.monkeypatch.setattr(dbedit_remove, "Session", self.Session)
        self.monkeypatch.setattr(
            dbedit_remove, "get_trial_if_exists", self.get_trial_if_exists
        )
        self.monkeypatch.setattr(
            dbedit_remove, "remove_clinical", self.mock_remove_clinical
        )
        self.monkeypatch.setattr(
            dbedit_remove, "DownloadableFiles", self.DownloadableFiles
        )
        self.monkeypatch.setattr(dbedit_remove, "TrialMetadata", self.TrialMetadata)
        self.monkeypatch.setattr("builtins.print", self.mock_print)

    def test_bail_outs(self):
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="bar", target_id=tuple()
        )
        self.mock_print.assert_called_once_with(
            "Assay / analysis not supported:", "bar"
        )

        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="clinical_data", target_id=("bar",)
        )
        self.mock_print.assert_not_called()
        self.mock_remove_clinical.assert_called_once_with(
            trial_id="foo", target_id="bar"
        )

    def test_olink(self):
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="olink", target_id=("bar",)
        )
        self.mock_print.assert_called_once_with(
            "Cannot find olink batch bar for trial foo"
        )
        self.query.assert_not_called()

        # if no combined file, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="olink",
            target_id=("olink_batch_2", "combined"),
        )
        self.mock_print.assert_called_once_with(
            "Cannot find a combined file for olink batch olink_batch_2 for trial foo"
        )
        self.query.assert_not_called()

        # if no matching file, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="olink",
            target_id=("olink_batch_2", "bar"),
        )
        self.mock_print.assert_called_once_with(
            "Cannot find a file bar in olink batch olink_batch_2 for trial foo"
        )
        self.query.assert_not_called()

        # remove a single file
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["assays"]["olink"]["batches"][1]["records"].pop(0)
        self.mock_print.reset_mock()
        target_id = (
            "olink_batch_2",
            f"{TEST_TRIAL_ID}/olink/batch_olink_batch_2/chip_0/assay_npx.xlsx",
        )
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="olink",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing olink values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/olink/batch_olink_batch_2/chip_0/assay_npx.xlsx",
            ]
        )

        # remove a combined file
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["assays"]["olink"]["batches"][0].pop("combined")
        self.mock_print.reset_mock()
        target_id = (
            "olink_batch",
            "combined",
        )
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="olink",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing olink values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/olink/batch_olink_batch/combined_npx.xlsx",
            ]
        )

        # remove a whole batch
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["assays"]["olink"]["batches"].pop(1)
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="olink",
            target_id=("olink_batch_2",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing olink values {('olink_batch_2',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/olink/batch_olink_batch_2/chip_1/assay_npx.xlsx",
            ]
        )

        # remove a study-wide file
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["assays"]["olink"].pop("study")
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="olink",
            target_id=("study",),
        )

        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing olink values {('study',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/olink/study_npx.xlsx",
                f"{TEST_TRIAL_ID}/olink/ALL object_urls",
            ]
        )

        # remove last record removes whole batch and assay
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["assays"].pop("olink")
        self.mock_print.reset_mock()
        target_id = (
            "olink_batch",
            f"{TEST_TRIAL_ID}/olink/batch_olink_batch/chip_0/assay_npx.xlsx",
        )
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="olink",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing olink values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/olink/batch_olink_batch/chip_0/assay_npx.xlsx",
                f"{TEST_TRIAL_ID}/olink/batch_olink_batch/chip_0/ALL object_urls",
                f"{TEST_TRIAL_ID}/olink/batch_olink_batch/ALL object_urls",
                f"{TEST_TRIAL_ID}/olink/ALL object_urls",
            ]
        )

    def test_elisa(self):
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="elisa", target_id=("bar",)
        )
        self.mock_print.assert_called_once_with(
            "Cannot find elisa batch bar for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # remove a single run
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["assays"]["elisa"].pop(0)
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="elisa",
            target_id=("elisa_batch",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing elisa values {('elisa_batch',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/elisa/elisa_batch/assay.xlsx",
            ]
        )

        # remove last run removes whole assay
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["assays"].pop("elisa")
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="elisa",
            target_id=("elisa_batch_2",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing elisa values {('elisa_batch_2',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/elisa/elisa_batch_2/assay.xlsx",
                f"{TEST_TRIAL_ID}/elisa/elisa_batch_2/ALL object_urls",
            ]
        )

    def test_nanostring(self):
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="nanostring", target_id=("bar",)
        )
        self.mock_print.assert_called_once_with(
            "Cannot find nanostring batch bar for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # if no matching run_id, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="nanostring",
            target_id=("nanostring_batch", "bar"),
        )
        self.mock_print.assert_called_once_with(
            "Cannot find a run bar in nanostring batch nanostring_batch for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # remove a single run
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["assays"]["nanostring"][0]["runs"].pop(0)
        self.mock_print.reset_mock()
        target_id = (
            "nanostring_batch",
            "nanostring_batch_run",
        )
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="nanostring",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing nanostring values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/nanostring_batch_run/control.rcc",
                f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/nanostring_batch_run/CTTTPP101.00.rcc",
            ]
        )

        # remove a single batch
        target_metadata["assays"]["nanostring"].pop(1)
        self.mock_print.reset_mock()
        self.DownloadableFiles.object_url.reset_mock()
        self.filter_query.update.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="nanostring",
            target_id=("nanostring_batch_2",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing nanostring values {('nanostring_batch_2',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/nanostring/nanostring_batch_2/normalized_data.csv",
                f"{TEST_TRIAL_ID}/nanostring/nanostring_batch_2/nanostring_batch_2_run/control.rcc",
                f"{TEST_TRIAL_ID}/nanostring/nanostring_batch_2/nanostring_batch_2_run/CTTTPP102.00.rcc",
            ]
        )
        # remove last run removes whole batch and assay
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["assays"].pop("nanostring")
        self.mock_print.reset_mock()
        target_id = ("nanostring_batch", "nanostring_batch_run_2")
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="nanostring",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing nanostring values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/nanostring_batch_run_2/control.rcc",
                f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/nanostring_batch_run_2/CTTTPP201.00.rcc",
                f"{TEST_TRIAL_ID}/nanostring/nanostring_batch/normalized_data.csv",
            ]
        )

    def test_cytof_analysis(self):
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="cytof_analysis", target_id=("bar",)
        )
        self.mock_print.assert_called_once_with(
            "Cannot find cytof analysis batch bar for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # if no matching run_id, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="cytof_analysis",
            target_id=("cytof_batch", "bar"),
        )
        self.mock_print.assert_called_once_with(
            "Cannot find cytof analysis for sample bar in batch cytof_batch for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # remove a single run
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["assays"]["cytof"][0]["records"][0].pop("output_files")
        self.mock_print.reset_mock()
        target_id = (
            "cytof_batch",
            "CTTTPP101.00",
        )
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="cytof_analysis",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing cytof_analysis values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch/CTTTPP101.00/assignment.csv",
                f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch/CTTTPP101.00/source.fcs",
            ]
        )

        # remove a single batch
        target_metadata["assays"]["cytof"][1].pop("astrolabe_analysis")
        target_metadata["assays"]["cytof"][1]["records"][0].pop("output_files")
        self.mock_print.reset_mock()
        self.DownloadableFiles.object_url.reset_mock()
        self.filter_query.update.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="cytof_analysis",
            target_id=("cytof_batch_2",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing cytof_analysis values {('cytof_batch_2',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch_2/reports.zip",
                f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch_2/CTTTPP102.00/assignment.csv",
                f"{TEST_TRIAL_ID}/cytof_analysis/cytof_run/cytof_batch_2/CTTTPP102.00/source.fcs",
            ]
        )

    def test_rna_level1_analysis(self):
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="rna_level1_analysis", target_id=("bar",)
        )
        self.mock_print.assert_called_once_with(
            "Cannot find RNA analysis for bar for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # remove a single sample
        self.mock_print.reset_mock()
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["analysis"]["rna_analysis"]["level_1"].pop(0)
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="rna_level1_analysis",
            target_id=("CTTTPP101.00",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing rna_level1_analysis values {('CTTTPP101.00',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/rna/CTTTPP101.00/analysis/error.yaml",
            ],
        )

        # remove down to a single samples
        self.mock_print.reset_mock()
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["analysis"]["rna_analysis"]["level_1"].pop(
            1
        )  # the last of 3 (now 2)
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="rna_level1_analysis",
            target_id=("CTTTPP102.00",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing rna_level1_analysis values {('CTTTPP102.00',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/rna/CTTTPP102.00/analysis/error.yaml",
            ],
        )

        # remove last sample removes whole assay
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        self.mock_print.reset_mock()
        target_metadata["analysis"].pop("rna_analysis")
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="rna_level1_analysis",
            target_id=("CTTTPP201.00",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing rna_level1_analysis values {('CTTTPP201.00',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/rna/CTTTPP201.00/analysis/error.yaml",
            ],
        )

    def test_wes_analysis(self):
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="wes_analysis", target_id=("bar",)
        )
        self.mock_print.assert_called_once_with(
            "Cannot find WES paired analysis for bar for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # remove a single pair run
        self.mock_print.reset_mock()
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["analysis"]["wes_analysis"]["pair_runs"].pop(0)
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes_analysis",
            target_id=("CTTTPP101.00",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing wes_analysis values {('CTTTPP101.00',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/wes/CTTTPP101.00/analysis/error.yaml",
            ],
        )

        # remove a pair run in both "wes_analysis" and "wes_analysis_old"
        self.mock_print.reset_mock()
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["analysis"]["wes_analysis"]["pair_runs"].pop(
            0
        )  # the second of 3, now first 2
        target_metadata["analysis"]["wes_analysis_old"]["pair_runs"].pop(1)
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes_analysis",
            target_id=("CTTTPP102.00",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing wes_analysis values {('CTTTPP102.00',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/wes/CTTTPP102.00/analysis/error.yaml",
                f"{TEST_TRIAL_ID}/wes/CTTTPP102.00/analysis/error.yaml",
            ],
        )

        # old_only, there's another one in wes_analysis
        # remove last pair run removes whole assay
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        self.mock_print.reset_mock()
        target_metadata["analysis"].pop("wes_analysis_old")
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes_analysis_old",
            target_id=("CTTTPP201.00",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing wes_analysis_old values {('CTTTPP201.00',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/wes/CTTTPP201.00/analysis/error.yaml",
            ],
        )

    def test_wes_tumor_only_analysis(self):
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes_tumor_only_analysis",
            target_id=("bar",),
        )
        self.mock_print.assert_called_once_with(
            "Cannot find WES tumor-only analysis for bar for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # remove a single pair run
        self.mock_print.reset_mock()
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["analysis"]["wes_tumor_only_analysis"]["runs"].pop(0)
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes_tumor_only_analysis",
            target_id=("CTTTPP101.00",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing wes_tumor_only_analysis values {('CTTTPP101.00',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/wes_tumor_only/CTTTPP101.00/analysis/error.yaml",
            ],
        )

        # remove a pair run in both "wes_tumor_only_analysis" and "wes_tumor_only_analysis_old"
        self.mock_print.reset_mock()
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["analysis"]["wes_tumor_only_analysis"]["runs"].pop(
            0
        )  # the second of 3, now first 2
        target_metadata["analysis"]["wes_tumor_only_analysis_old"]["runs"].pop(1)
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes_tumor_only_analysis",
            target_id=("CTTTPP102.00",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing wes_tumor_only_analysis values {('CTTTPP102.00',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/wes_tumor_only/CTTTPP102.00/analysis/error.yaml",
                f"{TEST_TRIAL_ID}/wes_tumor_only/CTTTPP102.00/analysis/error.yaml",
            ],
        )

        # old_only, there's another one in wes_tumor_only_analysis
        # remove last pair run removes whole assay
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        self.mock_print.reset_mock()
        target_metadata["analysis"].pop("wes_tumor_only_analysis_old")
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes_tumor_only_analysis_old",
            target_id=("CTTTPP201.00",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing wes_tumor_only_analysis_old values {('CTTTPP201.00',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/wes_tumor_only/CTTTPP201.00/analysis/error.yaml",
            ],
        )

    def test_batched_analysis(self):
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="tcr_analysis", target_id=("bar",)
        )
        self.mock_print.assert_called_once_with(
            "Cannot find tcr_analysis batch bar for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # if no matching cimac_id, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="tcr_analysis",
            target_id=("tcr_analysis_batch", "bar"),
        )
        self.mock_print.assert_called_once_with(
            "Cannot find tcr_analysis for sample bar in batch tcr_analysis_batch for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # remove a single batch
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["analysis"]["tcr_analysis"]["batches"][0]["records"].pop(0)
        self.mock_print.reset_mock()
        target_id = (
            "tcr_analysis_batch",
            "CTTTPP101.00",
        )
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="tcr_analysis",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing tcr_analysis values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch/CTTTPP101.00/tra_clone.csv",
            ]
        )

        # remove a single batch
        target_metadata["analysis"]["tcr_analysis"]["batches"].pop(1)
        self.mock_print.reset_mock()
        self.DownloadableFiles.object_url.reset_mock()
        self.filter_query.update.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="tcr_analysis",
            target_id=("tcr_analysis_batch_2",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing tcr_analysis values {('tcr_analysis_batch_2',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch_2/report_trial.tar.gz",
                f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch_2/CTTTPP102.00/tra_clone.csv",
            ]
        )
        # remove last run removes whole batch and assay
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["analysis"].pop("tcr_analysis")
        self.mock_print.reset_mock()
        target_id = ("tcr_analysis_batch", "CTTTPP201.00")
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="tcr_analysis",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing tcr_analysis values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch/CTTTPP201.00/tra_clone.csv",
                f"{TEST_TRIAL_ID}/tcr_analysis/tcr_analysis_batch/report_trial.tar.gz",
            ]
        )

    def test_generic_assay_with_ids(self):
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="atacseq_analysis", target_id=("bar",)
        )
        self.mock_print.assert_called_once_with(
            "Cannot find atacseq_analysis batch bar for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # if no matching cimac_id, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="atacseq_analysis",
            target_id=("atacseq_analysis_batch", "bar"),
        )
        self.mock_print.assert_called_once_with(
            "Cannot find atacseq_analysis for sample bar in batch atacseq_analysis_batch for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # remove a single batch
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["analysis"]["atacseq_analysis"][0]["records"].pop(0)
        self.mock_print.reset_mock()
        target_id = (
            "atacseq_analysis_batch",
            "CTTTPP101.00",
        )
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="atacseq_analysis",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing atacseq_analysis values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/atacseq/CTTTPP101.00/analysis/aligned_sorted.bam",
            ]
        )

        # remove a single batch
        target_metadata["analysis"]["atacseq_analysis"].pop(1)
        self.mock_print.reset_mock()
        self.DownloadableFiles.object_url.reset_mock()
        self.filter_query.update.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="atacseq_analysis",
            target_id=("atacseq_analysis_batch_2",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing atacseq_analysis values {('atacseq_analysis_batch_2',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/atacseq/analysis/atacseq_analysis_batch_2/report.zip",
                f"{TEST_TRIAL_ID}/atacseq/CTTTPP102.00/analysis/aligned_sorted.bam",
            ]
        )

        # remove last run removes whole batch and assay
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["analysis"].pop("atacseq_analysis")
        self.mock_print.reset_mock()
        target_id = ("atacseq_analysis_batch", "CTTTPP201.00")
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="atacseq_analysis",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing atacseq_analysis values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/atacseq/CTTTPP201.00/analysis/aligned_sorted.bam",
                f"{TEST_TRIAL_ID}/atacseq/analysis/atacseq_analysis_batch/report.zip",
            ]
        )

    def test_generic_assay_no_ids(self):
        self.mock_print.reset_mock()
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="wes", target_id=("bar",)
        )
        # if no matching batch, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo", assay_or_analysis="wes", target_id=("bar",)
        )
        self.mock_print.assert_called_once_with(
            "Cannot find wes batch bar for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # if no matching cimac_id, bails
        self.mock_print.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes",
            target_id=("0", "bar"),
        )
        self.mock_print.assert_called_once_with(
            "Cannot find wes for sample bar in batch 0 for trial foo"
        )
        self.query.assert_not_called()
        self.session.delete.assert_not_called()

        # remove a single batch
        target_metadata = deepcopy(TEST_METADATA_JSON)
        target_metadata["assays"]["wes"][0]["records"].pop(0)
        self.mock_print.reset_mock()
        target_id = (
            "0",
            "CTTTPP101.00",
        )
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing wes values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/wes/CTTTPP101.00/reads_0.bam",
                f"{TEST_TRIAL_ID}/wes/CTTTPP101.00/reads_1.bam",
            ]
        )

        # remove a single batch
        target_metadata["assays"]["wes"].pop(1)
        self.mock_print.reset_mock()
        self.DownloadableFiles.object_url.reset_mock()
        self.filter_query.update.reset_mock()
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes",
            target_id=("1",),
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing wes values {('1',)}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/wes/CTTTPP102.00/reads_0.bam",
            ]
        )
        # remove last run removes whole batch and assay
        self.filter_query.update.reset_mock()
        self.DownloadableFiles.object_url.in_.reset_mock()
        target_metadata["assays"].pop("wes")
        self.mock_print.reset_mock()
        target_id = ("0", "CTTTPP201.00")
        dbedit_remove.remove_data(
            trial_id="foo",
            assay_or_analysis="wes",
            target_id=target_id,
        )
        self.mock_print.assert_called_once_with(
            f"Updated trial foo, removing wes values {target_id}",
            f"along with {self.filter_query.delete()} files",
        )
        self.filter_query.update.assert_called_once()
        args, _ = self.filter_query.update.call_args
        assert (
            DeepDiff(args[0][self.TrialMetadata.metadata_json], target_metadata) == {}
        )
        self.DownloadableFiles.object_url.in_.assert_called_once_with(
            [
                f"{TEST_TRIAL_ID}/wes/CTTTPP201.00/reads_0.bam",
            ]
        )



from threading import Thread


class ExceptionCatchingThread(Thread):
    """
    A thread that throws any exception it encounters while
    running when its `join` method is called.
    """

    def __init__(self, target):
        Thread.__init__(self)
        self.target = target

    def run(self):
        self.exc = None
        try:
            self.target()
        except:
            import sys

            self.exc = sys.exc_info()

    def join(self):
        Thread.join(self)
        if self.exc:
            msg = "Thread '%s' threw an exception: %s" % (self.getName(), self.exc[1])
            new_exc = self.exc[0](msg)
            new_exc.with_traceback(self.exc[2])
            raise new_exc



#!/usr/bin/env python3
"""
Setup for the package
"""

from setuptools import setup, find_packages

from nci_cidc_cli import __version__

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setup(
    name="nci_cidc_cli",
    version=__version__,
    packages=find_packages(exclude=("tests")),
    entry_points={"console_scripts": ["nci-cidc = nci_cidc_cli.cli:cidc"]},
    description="A command-line interface for interacting with the NCI CIDC.",
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    python_requires=">=3.6",
)
